@@ -65,10 +65,9 @@ class data_field_latlong extends data_field_base {
         $classes = 'mod-data-input form-control-static';
         $str .= '<label for="field_'.$this->field->id.'_0" class="' . $classes . '">' . get_string('latitude', 'data');
         if ($this->field->required) {
-            $str .= html_writer::img($OUTPUT->pix_url('req'), get_string('requiredelement', 'form'),
-                                     array('class' => 'req', 'title' => get_string('requiredelement', 'form')));
+            $str .= $OUTPUT->pix_icon('req', get_string('requiredelement', 'form'));
         }
-        $classes = 'form-control m-x-1';
+        $classes = 'form-control mx-1';
         $str .= '</label></td><td>';
         $str .= '<input class="' . $classes . '" type="text" name="field_'.$this->field->id.'_0" ';
         $str .= ' id="field_'.$this->field->id.'_0" value="';
@@ -77,10 +76,9 @@ class data_field_latlong extends data_field_base {
         $str .= '<tr><td align="right"><label for="field_'.$this->field->id.'_1" class="' . $classes . '">';
         $str .= get_string('longitude', 'data');
         if ($this->field->required) {
-            $str .= html_writer::img($OUTPUT->pix_url('req'), get_string('requiredelement', 'form'),
-                                     array('class' => 'req', 'title' => get_string('requiredelement', 'form')));
+            $str .= $OUTPUT->pix_icon('req', get_string('requiredelement', 'form'));
         }
-        $classes = 'form-control m-x-1';
+        $classes = 'form-control mx-1';
         $str .= '</label></td><td><input class="' . $classes . '" type="text" ';
         $str .= 'name="field_'.$this->field->id.'_1" id="field_'.$this->field->id.'_1" value="';
         $str .= s($long).'" size="10" />°E</td>';
@@ -120,8 +118,12 @@ class data_field_latlong extends data_field_base {
        return $return;
     }
 
-    function parse_search_field() {
-        return optional_param('f_'.$this->field->id, '', PARAM_NOTAGS);
+    public function parse_search_field($defaults = null) {
+        $param = 'f_'.$this->field->id;
+        if (empty($defaults[$param])) {
+            $defaults = array($param => '');
+        }
+        return optional_param($param, $defaults[$param], PARAM_NOTAGS);
     }
 
     function generate_sql($tablealias, $value) {
@@ -197,7 +199,7 @@ class data_field_latlong extends data_field_base {
                 }
                 // NB! If you are editing this, make sure you don't break the javascript reference "previousSibling"
                 //   which allows the "Go" button to refer to the drop-down selector.
-                $str .= '\n</select><input type="button" class="btn m-l-1 btn-secondary" value="' . get_string('go');
+                $str .= '\n</select><input type="button" class="btn ml-1 btn-secondary" value="' . get_string('go');
                 $str .= '" onclick="if(previousSibling.value){self.location=previousSibling.value}"/>';
                 $str .= '</form>';
             } else {
@@ -296,4 +298,19 @@ class data_field_latlong extends data_field_base {
         // If we get here then only one field has been filled in.
         return get_string('latlongboth', 'data');
     }
+
+    /**
+     * Return the plugin configs for external functions.
+     *
+     * @return array the list of config parameters
+     * @since Moodle 3.3
+     */
+    public function get_config_for_external() {
+        // Return all the config parameters.
+        $configs = [];
+        for ($i = 1; $i <= 10; $i++) {
+            $configs["param$i"] = $this->field->{"param$i"};
+        }
+        return $configs;
+    }
 }
